


(* csp2B.sml version 1.2

Changes since 1.1 : Fixed bug in subst_ins which cause it
to lose some out/dot params.  Better layout of operations.

Changes since 1.0 :  Added preamble to machines and refinements
allowing USES, IMPORTS, etc

*)




structure csp2B :> csp2B  = 
struct


val symbols =   (map explode ( [ "->", "[]" , "|||", "<--" ,
"=>",
"<=>",
"/=",
"==",
"|->",
"/:",
"<:",
"/<:",
"\\/",
"/\\",
"{}",
">=",
"<=",
"..",
"<->",
"<|",
"|>",
"<<|",
"|>>",
"<+",
"+>",
"><",
"||",
"+->",
"-->",
">+>",
">->",
"+->>",
"-->>",
">->>",
"<>",
"->",
"<-",
"/|\\",
"\\|/"
] ) ) 
;


val exists = List.exists
val map = List.map
val filter = List.filter
val length = List.length
val foldr = List.foldr
val drop = List.drop

val mkDict = Binarymap.mkDict
val insert = Binarymap.insert
val listItems = Binarymap.listItems
val peek = Binarymap.peek
val find = Binarymap.find
val remove = Binarymap.remove

val member = Binaryset.member
val add = Binaryset.add
val addList = Binaryset.addList
val empty = Binaryset.empty

val InttoString =  Int.toString

val concat = String.concat 
val scompare =  String.compare;

val isAlpha = Char.isAlpha
val isDigit = Char.isDigit
val isAlphaNum = Char.isAlphaNum
val toLower = Char.toLower

val zip = ListPair.zip
val unzip = ListPair.unzip


val DatetoString = Date.toString
val fromTimeLocal = Date.fromTimeLocal
val now = Time.now

val openOut = TextIO.openOut
val closeOut = TextIO.closeOut
val print = TextIO.print
val output = TextIO.output
val openIn = TextIO.openIn
val inputAll = TextIO.inputAll
val closeIn = TextIO.closeIn

val arguments = CommandLine.arguments


type IDENT = string;

type EXPR = string list;

type TYPE = string ;

datatype PARAM = In of IDENT
               | Out of EXPR
               | Dot of EXPR;

datatype EVENT = Event of ( IDENT * (PARAM list) );

datatype PROC = Stop
              | Prefix of (EXPR * EVENT * PROC)
              | Sum of (PROC list)
              | Cond of (EXPR * PROC)
              | Recurse of (IDENT * (EXPR list)) ;

type LHS = IDENT * (IDENT list);

type EQN = (LHS * PROC);

datatype ENV = Env of (
        (IDENT, TYPE) Binarymap.dict *       (* types *)
        (IDENT list) *      (* indices *)
        (IDENT, (IDENT list) * (IDENT list) * (IDENT list)) Binarymap.dict (* events *)
                       ) ;

type LexSource = int * int * (char list) ;













fun fst (x,y) = x;
fun snd (x,y) = y;

infix mem;
fun x mem [] = false
  | x mem (y::ys) = (x=y) orelse x mem ys;


                
fun is_alpha c =    isAlpha c orelse c= #"_" ;

val is_num  =  isDigit  ;

val is_alphanum  = isAlphaNum;
                                        


fun front [] = []
  | front [x] = []
  | front (x::(y::ys)) = x::(front (y::ys));

(*
fun input_symbols infile =
    let val s = TextIO.inputLine infile
    in
        if s = "" then []
        else 
        (front (explode s))::(input_symbols infile)
    end
;

fun get_bsymbols () = 
  let 
    val infile = TextIO.openIn "Bsymbols"
    val m = input_symbols infile
    val _ = TextIO.closeIn infile
  in
       m
  end 
; 



val symbols =   (map explode ( [ "->", "[]" , "|||", "<--" ] ) ) @
                ( get_bsymbols() )
;
*)


fun strict_prefixes [] = []
  | strict_prefixes [c] = []
  | strict_prefixes (c1::(c2::cs)) =
      let val p = front (c1::(c2::cs))
      in   p::(strict_prefixes p) end;

        

fun rc_foldr f id [] = id
  | rc_foldr f id (x::xs) = f (x, (rc_foldr f id xs)) ;

fun prefixes syms= rc_foldr op@ [] (map strict_prefixes syms) ;

val symbol_prefixes =   ( prefixes symbols ) ;




fun scan_ident id i j []  = (implode id,(i,j,[]))
  | scan_ident id i j (c::cs) =
                if is_alphanum c
                then scan_ident (id@[c]) i (j+1) cs
                else (implode id, (i,j,(c::cs)));



fun drop_last [x] = ([],x)
  | drop_last (x::xs) =
        let val (ys,y) = drop_last xs in
        ((x::ys),y) end;

fun symbol_backtrack sy i j cs =
        let val (sy2,c) = drop_last sy in
        if sy2 = []
        then (str c,(i,j,cs))
        else if sy2 mem symbols
        then (implode sy2, (i, j-1, (c::cs) ) )
        else symbol_backtrack sy2 i (j-1) (c::cs)
        end;



fun scan_symbol sy i j [] =
                symbol_backtrack sy i j []
  | scan_symbol sy i j (c::cs) =
                let val sy2 = sy@[c] in
                        if sy2 mem (symbol_prefixes)
                        then scan_symbol sy2 i (j+1) cs
                        else if sy2 mem (symbols) 
                        then (implode sy2, (i, j+1, cs))
                        else symbol_backtrack sy2 i (j+1) cs
                end ;



fun 
    scan_comment i j [] = ("",(i,j,[]))
  | scan_comment i j [c1] = ("",(i,j+1,[]))
  | scan_comment i j (c1::c2::cs) =
        if c1= #"*" andalso c2= #"/"
        then lexscan (i,(j+2),cs)
        else if c1 = #"\n"
        then scan_comment (i+1) 0 (c2::cs)
        else scan_comment i (j+1) (c2::cs)
and
        lexscan (i,j,[]) = ("",(i,j,[]))
  |     lexscan (i,j,(c::cs)) =
                if is_alpha  c
                then (* identifier *)
                        scan_ident [c] i (j+1) cs
                else if c = #"/" andalso not(cs=[]) andalso ((hd cs)= #"*")
                then (* comment *)
                        scan_comment i (j+2) (tl cs)
                else if [c] mem (symbol_prefixes)
                then (* maybe a special symbol *)
                        scan_symbol [c] i (j+1) cs 
                else if c = #" " orelse c = #"\t"
                then (* skip spaces *)
                        lexscan (i, (j+1), cs)
                else if c = #"\n"
                then (* skip newlines *)
                        lexscan ((i+1), 0, cs)
                else (* other token *)
                        ( str c, (i,j+1,cs) ) 
;


fun match_keyword _ [] = true
  | match_keyword [] (w::ws) = false
  | match_keyword (c::cs) (w::ws) = 
        c=w andalso match_keyword cs ws
;


fun 
    rem_comment2 i j [] = (i,j,[])
  | rem_comment2 i j [c1] = (i,j,[c1])
  | rem_comment2 i j (c1::c2::cs) =
        if c1= #"*" andalso c2= #"/"
        then (i,(j+2),cs)
        else if c1 = #"\n"
        then rem_comment2 (i+1) 0 (c2::cs)
        else rem_comment2 i (j+1) (c2::cs)
;

fun 
    rem_comment i j [] = (i,j,[])
  | rem_comment i j [c1] = (i,j,[c1])
  | rem_comment i j (c1::c2::cs) =
        if c1= #"/" andalso c2= #"*"
        then rem_comment2 i (j+2) cs
        else (i, j, (c1::c2::cs))
;


fun     lexscan_until2 res _ (i,j,[]) = (res,("",(i,j,[])))
  |     lexscan_until2 res ids (i,j,(c::cs)) =
            let val (i2,j2,cs2) = rem_comment i j (c::cs)
            in
              if cs2 = [] 
              then (res,("",(i2,j2,[])))
              else if  exists (match_keyword cs2) ids
              then (res, lexscan (i2,j2,cs2) )
              else let val (i3,j3) =  (if (hd cs2) = #"\n" 
                                    then (i2+1,0)
                                    else (i2,j2+1) )
                 in
                    lexscan_until2 (res ^ (str (hd cs2) )) ids (i3,j3,tl cs2)
                 end
            end
;



fun lexscan_until ids r = lexscan_until2 "" (map explode ids) r;


fun init_string s = lexscan (1,0,(explode s));

fun init_string2 s =  (1,0,(explode s));

fun get_pos (t,(i,j,s)) =  (i,j-(size t));















































fun pairlist_to_bmap comp [] = 
        mkDict comp
  | pairlist_to_bmap comp ((x,y)::ps) =
        insert(pairlist_to_bmap comp ps,x,y)
;


fun pcompare ((a1,a2),(b1,b2)) =
    case scompare (a1,b1) of
        LESS => LESS
      | GREATER => GREATER
      | EQUAL => scompare (a2,b2)
;

fun listpair_to_bmap comp x = pairlist_to_bmap comp (zip x);


fun MK_env chans chantypes indices eventdecs  = 
    let fun f i j= find(chantypes,(i,j))
        fun g (id,ins,outs) = (id,(ins,map (f id) ins,outs))
        val eventinfo = pairlist_to_bmap scompare (map g eventdecs)
        val (eid,_,_) = hd eventdecs
        fun h i = find(chantypes,(eid,i))
        val types = listpair_to_bmap scompare (indices, map h indices)
    in
        Env (types,indices,eventinfo)
    end
;


fun MK_chantypes chans =
    let fun g m cid ins [] = m
          | g m cid [] _ = m
          | g m cid (i::iis) (j::jjs) = g (insert(m,(cid,i),j)) cid iis jjs
        fun f m [] =  m 
          | f m ((cid,(ins,intyps),_)::t) = f (g m cid ins intyps) t
    in
        f (mkDict pcompare) chans
    end
;


fun get_types (Env(types,_,_)) = types
;

fun get_var_ids (Env(types,indices,_)) =
    let fun f k = not(k mem indices)
    in
        filter f (map fst (listItems types))
        end
;



fun MK_update_types env ([],_) = env
  | MK_update_types env (_,[]) = env
  | MK_update_types (Env(a,b,c)) (v::vs,t::ts) =
        if t="?"
        then case peek(a,v) of
                (SOME _) => MK_update_types (Env(a,b,c)) (vs,ts)
               | NONE => let val a2 = insert(a,v,t)
                         in MK_update_types (Env(a2,b,c)) (vs,ts)
                         end
        else let val a2 = insert(a,v,t)
             in MK_update_types (Env(a2,b,c)) (vs,ts)
             end
;


val keywords =
    [ "MACHINE", "REFINEMENT", "ALPHABET",
    "INCLUDES", "WHERE", "END", "STOP", "INVARIANT",
    "CONJOINS", "REFINES", "PROCESS", "CONSTRAINS",
    "DEFINITIONS", "SETS" ] ;
                                      
fun is_identifier s = 
    not(s="") andalso 
    isAlpha(hd (explode s)) andalso 
    not( s mem keywords );


fun is_eventname (Env (_,_,es)) e  =  
    case peek(es,e) of
        NONE => false
      | (SOME _) => true
;


fun get_param_lengths (Env (_,_,es)) eid =
    let val (l1,_,l2) = find(es,eid)
    in (length l1, length l2) 
    end
;


fun get_inchan_info (Env(_,_,m)) eid  =
    let val (ins,types,_) =  find(m,eid)  
    in (ins,types)
    end
;


fun add_args env _ [] = env
  | add_args env [] _ = env
  | add_args (Env(m,a,b)) (v::vs) (t::ts) =
        let val m2 = insert(m,v,t)
            val env2 = Env(m2,a,b)
        in
            add_args env2 vs ts
        end
;


fun get_indices (Env(_,i,_)) = i; 


fun PP_listop opr [] = ""
  | PP_listop opr [x] = x
  | PP_listop opr (x::y::t) = 
        if x = "" 
        then (PP_listop opr (y::t))
        else if y=""
             then (PP_listop opr (x::t))
             else x ^ opr ^ (PP_listop opr (y::t))
;


fun expr_to_string e = PP_listop (" ")  e;


fun ident_to_expr id = [id];


fun MK_and e1 e2 = ["("] @ e1 @ [") & ("] @ e2 @ [")"]; 

fun MK_not e = ["not("] @ e @ [")"];

fun MK_inparam id = In id;

fun MK_outparam exp = Out exp;

fun MK_dotparam exp = Dot exp;


fun MK_event id params = (Event (id,params));

val MK_stop = Stop;

fun MK_sum [] = Sum []
  | MK_sum [p] = p
  | MK_sum (p::(q::qs)) =  Sum (p::(q::qs));
  
fun MK_prefix ev proc = Prefix ([],ev,proc); 

fun MK_if ex proc = Cond (ex,proc); 

fun MK_ifelse g p q  =  Sum [ Cond(g,p), Cond(MK_not g, q) ] ;

fun MK_recurse id args =
     Recurse (id,args) 
    ;


fun MK_lhs id vars = (id,vars);

fun MK_eqn lhs proc =  (lhs,proc);



val MK_empty_fresh = 
    (mkDict scompare):(string,int)Binarymap.dict;

fun MK_freshid fresh id =
    case  peek(fresh,id)  of
        NONE => ( insert(fresh,id,2), id ^ "_1" )
      | SOME c => ( insert(fresh,id,c+1), id ^ "_" ^(InttoString c) )
;


fun MK_freshidlist fr ids [] = (fr,ids)
  | MK_freshidlist fr ids ( id::t ) = 
        let val (fr2,id2) = MK_freshid fr id
        in  MK_freshidlist fr2 (ids@[id2]) t  end
;



fun MK_substmap [] = 
        (mkDict scompare):(string,string)Binarymap.dict
  | MK_substmap ((v,e)::t) =
        insert( MK_substmap t, v, e )
;


fun subst_ident sub id =
    case peek(sub,id) of
        NONE => id
      | SOME x => x
;

fun subst_expr sub e = map (subst_ident sub) e
;



fun remove_all sub [] = sub
  | remove_all sub (k::ks) =
     case peek(sub,k) of
        NONE => remove_all sub ks
      | SOME _ => remove_all (fst(remove(sub,k))) ks
;



fun subst_params sub [] = []
  | subst_params sub ( (In id)::t ) = 
        ( (In id)::subst_params sub t )
  | subst_params sub ( (Out e)::t ) = 
        ( (Out (subst_expr sub e))::subst_params sub t )
  | subst_params sub ( (Dot e)::t ) = 
        ( (Dot (subst_expr sub e))::subst_params sub t )
;



fun subst_ins _ [] = []
  | subst_ins [] t = t
  | subst_ins (id1::ids) ( (In id2)::t ) = 
        ( (In id1)::subst_ins ids t )
  | subst_ins ids ( (Out e)::t ) = 
        ( (Out e)::subst_ins ids t )
  | subst_ins ids ( (Dot e)::t ) = 
        ( (Dot e)::subst_ins ids t )
;



fun get_inparams [] = []
  | get_inparams ( (In id)::t ) = id::(get_inparams t)
  | get_inparams ( (Out _)::t ) = (get_inparams t)
  | get_inparams ( (Dot _)::t ) = (get_inparams t)
;

    

fun subst_proc sub Stop = Stop
  | subst_proc sub (Prefix(g,Event(eid,params),p)) = 
        let val ins = get_inparams params
            val sub2 = remove_all sub ins
            val params2 = subst_params sub2 params
            val p2 = subst_proc sub2 p
            val g2 = subst_expr sub g
        in (Prefix(g2,Event(eid,params2),p2))
        end
  | subst_proc sub (Sum ps) =
        Sum (map (subst_proc sub) ps)
  | subst_proc sub (Recurse(pid,exprs)) =
        Recurse(pid,map (subst_expr sub) exprs)
;
        


fun push_guard1 g Stop = Stop
  | push_guard1 g (Prefix(_,eid,p)) = Prefix(g,eid,push_guard p)
  | push_guard1 g (Sum ps) =
        Sum (map (push_guard1 g) ps)
  | push_guard1 g (Cond(h,p)) =
        push_guard1 (MK_and g h) p
and
    push_guard  Stop = Stop
  | push_guard (Recurse(x)) = Recurse(x)
  | push_guard (Prefix(g,ev,p)) = Prefix(g,ev,push_guard p)
  | push_guard (Sum ps) =
        Sum (map push_guard ps)
  | push_guard (Cond(g,p)) =
        push_guard1 g p
;



fun get_inparams2 [] _ = ([],[])
  | get_inparams2 _ [] = ([],[])
  | get_inparams2 ( (In id)::t ) (a::u) = 
        let val (l,m) = get_inparams2 t u
        in  (id::l, a::m)
        end
  | get_inparams2 ( (Out _)::t ) (a::u) = (get_inparams2 t u)
  | get_inparams2 ( (Dot _)::t ) (a::u) = (get_inparams2 t u)
;

    





fun 
    norm env fr pid lhs Stop = 
        ( env, fr, [ (lhs,Stop) ] )
  | norm env fr pid lhs ( Prefix(g,a,p) ) =
        ( case p of
            Stop => ( env, fr, [ (lhs,Prefix(g,a,p)) ] )
          | Recurse(_) => ( env, fr, [ (lhs,Prefix(g,a,p)) ] )
          | v => 
            let val (_,args) = lhs
                val Event(eid,params) = a
                val (in_names,in_types) = get_inchan_info env eid 
                val (in_binds,in_types2) =  (get_inparams2 params in_types)
                val (fr2,in_binds2) = MK_freshidlist fr [] in_binds
                val env2 = add_args env in_binds2 in_types2
                val params2 = subst_ins in_binds2 params
                val a2 = Event(eid,params2)
                val sub = MK_substmap (zip (in_binds,in_binds2) )
                val p2 = subst_proc sub p
                val args2 = (map (subst_ident sub) args) @ in_binds2
                val args3 = map ident_to_expr args2
                val (fr3,pid3) = MK_freshid fr2 pid
                val (env3,fr4, eqns2) = norm env2 fr3 pid (pid3,args2) p2
             in 
                ( env3, fr4,
                  (lhs, Prefix(g,a2,Recurse (pid3,args3))) :: eqns2 
                )
             end
        )
  | norm env fr pid lhs (Sum ps) = sumnorm env fr pid lhs [] ps
and
    sumnorm env  fr pid lhs eqns [] = (env,fr,eqns)
  | sumnorm env  fr pid lhs eqns (p::ps) = 
        let 
            val (env2,fr2,eqns2) = norm env fr pid lhs p
        in
            sumnorm env2 fr2 pid lhs (eqns@eqns2) ps
        end
;


fun eqnsnorm2 env fr eqns [] = (fr,env,eqns)
  | eqnsnorm2 env fr eqns1 (((pid,args),p)::eqns2) =
        let val (env2,fr2,eqns3) = norm env fr pid (pid,args) (push_guard p)
        in (eqnsnorm2 env2 fr2 (eqns1@eqns3) eqns2)
        end;


fun eqnsnorm env fresh eqns = eqnsnorm2 env fresh [] eqns;



fun match_dots sub [] _ = []
  | match_dots sub (x::xs) [] = []
  | match_dots sub (x::xs) (p::ps) =
    case p of
        (Dot e) => (x, expr_to_string(subst_expr sub e))::match_dots sub xs ps
      | v => match_dots sub xs ps
;


fun match_ins [] _ = []
  | match_ins (x::xs) [] = []
  | match_ins (x::xs) (p::ps) =
    case p of
        (In y) => (y,x)::match_ins xs ps
      | v => match_ins xs ps
;


fun get_outs [] = []
  | get_outs (p::ps) =
      case p of
        (Out e) => e::(get_outs ps)
      | v => get_outs ps
;





fun rem_redundancy1  [] = []
  | rem_redundancy1  ((x,y)::t) = 
        if y = "_"
        then rem_redundancy1 t
        else (x,y)::(rem_redundancy1 t)
;



fun rem_redundancy2  ([],_) = ([],[])
  | rem_redundancy2  (_,[]) = ([],[])
  | rem_redundancy2  (x::s, y::t) = 
        let val (s2,t2) = rem_redundancy2 (s,t)
        in
            if y =  "_"
            then (s2,t2)
            else (x::s2, y::t2)
        end
;

fun rem_dups_from_ass  ([],[]) = ([],[])
  | rem_dups_from_ass  ([],e::es) = 
        let val (xs2,es2) = rem_dups_from_ass ([],es)
        in
            ("??"::xs2, e::es2)
        end
  | rem_dups_from_ass  (x::xs, []) = 
        let val (xs2,es2) = rem_dups_from_ass (xs,[])
        in
            (x::xs2, "??"::es2)
        end
  | rem_dups_from_ass  (x::xs, e::es) = 
        let val (xs2,es2) = rem_dups_from_ass (xs,es)
        in
            if x = e  orelse  e =  "_"
            then (xs2,es2)
            else (x::xs2, e::es2)
        end
;


(*

type GUARD = (IDENT*EXPR)list;
type ASSIGN = (IDENT*EXPR)list;

type ACTION = EXPR * ControlGUARD * ParamGuard * 
              OutASSIGN * ControlASSIGN * ParamASSIGN;

*)



fun MK_action (vsub,cv,stop,pargs,ins,outs)
    ( (pid,args), Prefix( g, Event(_,params), Recurse(qid,exps) ) ) =
 let
    val cg = rem_redundancy1 [ (cv, pid) ]
    val pg = rem_redundancy1 (match_dots vsub ins params)
    val insub = MK_substmap (match_ins ins params)
    val outs2 = map (subst_expr insub) (get_outs params)
    val outs3 = map (subst_expr vsub) outs2
    val outs4 = map expr_to_string outs3
    val oa = rem_redundancy2 (outs,outs4)
    val exps2 = map (subst_expr insub) exps
    val exps3 = map (subst_expr vsub) exps2
    val exps4 = map expr_to_string exps3
    val lref1 = peek(pargs,qid)
    val (lref2,qid2) =
                case lref1 of
                    NONE  =>  ([],"Unknown Process '" ^ qid ^"'")
                  | (SOME l) => (l,qid)
    val lref3 = map (subst_ident vsub) lref2
    val pa = rem_dups_from_ass (lref3, exps4 )
    val ca = if pid=qid2 then ([],[]) else ([cv],[qid2])
    val g2  = (subst_expr vsub g,cg,pg)
    val a  = (oa,ca,pa)
 in
    (g2,a)
 end
  | MK_action (vsub,cv,stop,pargs,ins,outs)
    ( (pid,args), Prefix( g, Event(_,params), Stop ) ) =
 let
    val cg = rem_redundancy1 [ (cv, pid) ]
    val pg = rem_redundancy1 (match_dots vsub ins params)
    val insub = MK_substmap (match_ins ins params)
    val outs2 = map (subst_expr insub) (get_outs params)
    val outs3 = map (subst_expr vsub) outs2
    val outs4 = map expr_to_string outs3
    val oa = rem_redundancy2 (outs,outs4)
    val ca = ([cv],[stop])
    val g2  = (subst_expr vsub g,cg,pg)
    val a  = (oa,ca, ([],[]) )
 in
    (g2,a)
 end;
    

(* info =  var sub for indexed vars * indices *  
         control var * (Pid->args) * formal ins * formal outs *)

fun MK_actionlist info eqns eid1 =
    let fun f (_, Prefix(_,Event(eid2,_),_) ) = (eid1=eid2) 
          | f (_,Stop) = false
    in
        map (MK_action info) (filter f eqns)
    end;
    





fun PP_list [] = ""
  | PP_list [x] = x
  | PP_list (x::(y::ys)) = x ^ "," ^ (PP_list (y::ys));



fun PP_eql (x,y) =  x ^ " = " ^ y ;

fun tab i = if i = 0 then "" else "  " ^ (tab (i-1));




fun gd_expr_to_string i [] = ""
  | gd_expr_to_string i [x] = x
  | gd_expr_to_string i (x::(y::t)) = 
        if x = "&" 
        then "  &\n" ^ (tab i) ^ (gd_expr_to_string i (y::t))
        else x ^ " " ^ (gd_expr_to_string i (y::t))
;



fun PP_guard i (g,cg,pg) = 
    let val g_str = gd_expr_to_string i g
        val cg_str = if cg = [] 
                        then "" 
                        else PP_eql (hd cg)
        val pg_str = PP_listop (" & ") (map PP_eql pg)
        val guard = (tab i) ^ 
                    PP_listop ("  &\n" ^ (tab i) ) [ g_str, cg_str , pg_str ] ^
                    "\n"
    in
        guard
    end;


fun PP_assign (vs,es) = 
    if vs = []
    then "" 
    else (PP_list vs) ^ " := " ^ (PP_list es )
;



fun PP_body i (oa,ca,pa) = 
    let val oa_str = PP_assign oa
        val ca_str = PP_assign ca
        val pa_str = PP_assign pa
        val s = PP_listop ("  ||\n" ^ (tab i) ) [  ca_str, pa_str, oa_str ]
    in
        (tab i) ^ (if s = "" then "skip" else s) ^ "\n"
    end;




fun PP_actionlist2 i [] = (tab i) ^ "END"
  | PP_actionlist2 i ((g,b)::t) =
        (tab i) ^ "WHEN\n" ^ (PP_guard (i+2) g) ^ 
        (tab (i+1)) ^ "THEN\n" ^ (PP_body (i+2) b) ^ 
        (PP_actionlist2 i t)
;
        


fun PP_actionlist i [] = (tab i) ^ "SELECT not true THEN skip END"
  | PP_actionlist i ((g,b)::t) =
        (tab i) ^ "SELECT\n" ^ (PP_guard (i+2) g) ^ 
        (tab (i+1)) ^ "THEN\n" ^ (PP_body (i+2) b) ^ 
        (PP_actionlist2 i t)
;



fun PP_action indent (vsub,cv,stop,pargs,eqns) (eid,ins,outs) =
        let val act = MK_actionlist (vsub,cv,stop,pargs,ins,outs) eqns eid
            val gd_list = map (PP_guard 2) (map fst act)
            val act_str = PP_actionlist indent act
        in
                (eid, (gd_list, act_str, outs) )
        end
;




fun MK_pargs eqns  = pairlist_to_bmap scompare (map fst eqns);



fun rem_dups a [] = a
  | rem_dups a (x::s) =
        if x mem a 
        then rem_dups a s 
        else rem_dups (a@[x]) s
;


fun PP_controltype pid eqns stop = 
    let val l = (rem_dups [] (map (fst o fst) eqns))
        fun f (_,p) =
              case p of
                (Prefix(_,_,Stop)) =>  true
               | _ => false
        val l2 = if exists f eqns
                 then l @ [ stop ]
                 else l
        val s = PP_list  l2
    in  
        (tab 1) ^ pid ^ "State =\n" ^
        (tab 2) ^ "{" ^ s ^ "}"
    end
;


fun MK_varsubst indices types =
    let fun f (k,_) = not(k mem indices)
        val l1 = filter f (listItems types)
        fun g (id,_) =  (id, id ^ "(" ^ (PP_listop "|->" indices) ^ ")" )
    in
        if indices = [] 
        then  (l1, MK_substmap [] )
        else  (l1, MK_substmap (map g l1) )
    end
;


fun PP_var_inv indices types pid vars =
    let fun f id = find(types,id)
        val dtypes = map f indices
        val dom = 
            case dtypes of
                []  => ""
              | [x] => x ^ " --> "
              | v   => "(" ^ (PP_listop " * " dtypes) ^ ") --> "
        val cstate = pid ^ " : " ^ dom ^ pid ^ "State"
        fun g (x,t) =  x ^ " : " ^ dom ^ t
        val vstate = map g vars
    in
        (tab 1) ^ 
        (PP_listop (" &\n" ^ (tab 1)) (cstate::vstate) ) ^ "\n"
    end
;


fun PP_lambda params constraint e = 
        if params = []
        then e
        else "%(" ^ (PP_list params) ^ ").(" ^
             (PP_listop " & " constraint) ^ " | " ^ e ^ ")"
;


fun rem l [] = []
  | rem l ((x,y)::t) = if x mem l
                   then rem l t
                   else (x,y)::(rem l t)
;


fun PP_init pid initpid initargs indices types pargs pvars =
    let 
       val (vars,initpid2) =
                case peek(pargs,initpid) of
                    NONE  =>  ([],"Unknown Process '" ^ initpid ^"'")
                  | (SOME l) => (l,initpid)
        val (vars2,inits2) = rem_redundancy2 (vars,map expr_to_string initargs)
        val indtys = map (fn id => find(types,id)) indices
        val indplustys = map (fn (x,y) => (x ^ ":" ^ y))  (zip (indices,indtys))
        val inits3 = map (PP_lambda indices indplustys) inits2
        val s = PP_assign (vars2,inits3)
        val vars3 = rem vars2 pvars
        val (varlist,typelist) = unzip vars3
        val newvarlist = map (fn x => "new_" ^ x) varlist
        val dom = 
            case indtys of
                []  => ""
              | [x] => x ^ " --> "
              | v   => "(" ^ (PP_listop " * " indtys) ^ ") --> "
        val vtypes = map (fn (x,t) => x ^ " : " ^ dom ^ t)  (zip (newvarlist,typelist))
        val t = if varlist = []
                then ""
                else "ANY " ^ (PP_list newvarlist) ^ " WHERE\n" ^
                     (tab 2) ^ (PP_listop (" &\n" ^ (tab 2)) vtypes) ^
                     "\n" ^ (tab 1) ^ "THEN " ^
                     (PP_assign (varlist,newvarlist)) ^ " END"
        val r =  (PP_assign ( [pid], [PP_lambda indices indplustys initpid2] ) )
    in 
       (tab 1) ^ (PP_listop ("\n" ^ (tab 1) ^ "||\n" ^ (tab 1)) [r,s,t] ) ^ "\n"
   end
;



fun PP_process (pid,initpid,initargs,cargs,indices,types,eqns) =
    let 
        val pargs = MK_pargs eqns
        val cv =    pid ^ 
                    (if indices = [] then "" else "(" ^ (PP_listop "|->" indices) ^ ")")
        val (pvars,vsub) = MK_varsubst indices types
        val pvarlist = (map fst pvars)
        val stop = "Stop" ^ pid
    in 
        ( PP_controltype pid eqns stop,
          PP_list (pid::pvarlist),
          PP_var_inv indices types pid pvars,
          PP_init pid initpid initargs indices types pargs pvars,
          pairlist_to_bmap scompare (map (PP_action 3 (vsub,cv,stop,pargs,eqns)) cargs)
        )
    end
;








fun PP_outvars [] = ""
  | PP_outvars [x] = x ^ " <-- "
  | PP_outvars (x::y::T) = "(" ^ (PP_list (x::y::T)) ^ ") <-- "
;

fun PP_invars [] = ""
  | PP_invars (x::T) = "(" ^ (PP_list (x::T)) ^ ")"
;

fun PP_intypes ins =
    let fun f (x,y) = x ^ ":" ^ y in
        PP_listop " & " (map f (zip ins))
    end
;


fun get_actions actlist eid =
    let fun f NONE = false
          | f (SOME _) = true
        fun g m = peek(m,eid)
        fun h (SOME (x,y,z)) = (y,z)
    in 
        unzip (map h (filter f (map g actlist)))
    end
;


fun PP_operation mid inc actlist (eid,ins,outs) =
    let val (acts,outasslist) = get_actions actlist eid
        fun g id l = id mem l
        fun f id = not (exists (g id) outasslist)
        val callouts = PP_outvars (filter f (fst outs))
        val call = if inc
                   then if acts = []
                   then (tab 3) ^ callouts ^ eid ^ "_Act" ^ 
                        (PP_invars (fst ins))
                   else (tab 3) ^ "SELECT grd_" ^ mid ^ "_" ^ eid ^ " THEN " ^ 
                        callouts ^ eid ^ "_Act" ^ 
                        (PP_invars (fst ins)) ^ " END"
                   else ""
        val a_string = PP_listop ("\n" ^ (tab 2) ^ "||\n") (call::acts)
        val ty_string = PP_intypes ins
        val body = if a_string = ""
                   then (tab 1) ^ "skip"
                   else if ty_string = ""
                   then (tab 1) ^ "BEGIN\n" ^a_string ^ 
                        "\n" ^ (tab 1) ^ "END"
                   else (tab 1) ^ "PRE " ^ ty_string ^
                        " THEN\n" ^a_string ^ "\n" ^ (tab 1) ^
                        "END"
    in
        (PP_outvars (fst outs)) ^ 
        eid ^  (PP_invars (fst ins)) ^ " =\n" ^ body
    end
;
        



fun PP_operations mid inc chans actlist =
    PP_listop " ;\n\n" (map (PP_operation mid inc actlist) chans)
;
    



fun get_guards actlist eid =
    let fun f NONE = false
          | f (SOME _) = true
        fun g m = peek(m,eid)
        fun h (SOME (x,y,z)) = x
    in 
         (map h (filter f (map g actlist)))
    end
;



fun PP_guard_def  mid actlist (eid,ins,outs) =
    let val gdlist = get_guards actlist eid
    in if gdlist = []
       then ""
       else 
        let val gdlist2 = map (fn x => map (fn y => "(" ^ y ^ ")") x ) gdlist
            fun f [] = "not true"
              | f [e] = e
              | f (e::f::s) =  "(" ^ 
                    (PP_listop (" or\n" ^ (tab 1) ^ " ") (e::f::s) ) ^ ")"
            val gd = if length(gdlist2) = 1
                     then (tab 1) ^ (f (hd gdlist2))
                     else (tab 1) ^ "(\n" ^ (tab 1) ^
                           PP_listop ("\n" ^ (tab 1) ^ "& \n" ^ (tab 1) ) 
                                (map f gdlist2) ^ "\n" ^ (tab 1) ^ ")"
        in
           "grd_" ^ mid ^ "_" ^ eid ^ " == \n" ^ gd 
        end
    end
;
        


fun PP_guard_defs mid inc chans actlist =
        if inc
        then 
                PP_listop " ;\n\n" (map (PP_guard_def mid actlist) chans)
        else  ""
;
    





fun comp1 (x,_,_,_,_) = x;
fun comp2 (_,x,_,_,_) = x;
fun comp3 (_,_,x,_,_) = x;
fun comp4 (_,_,_,x,_) = x;
fun comp5 (_,_,_,_,x) = x;

fun PP_defs defs = if defs = ""
              then ""
              else "\n\nDEFINITIONS\n" ^ defs ^ "\n";


fun PP_rest mid (sets,defs,inv) inc chans pinfo_list =
    let val stuff = map PP_process pinfo_list
        val sets2 = PP_listop " ;\n" (map comp1 stuff)
        val vars = (tab 1) ^ PP_listop (",\n" ^ (tab 1)) (map comp2 stuff)
        val inv2 = PP_listop "  &\n" (map comp3 stuff)
        val init = PP_listop "  ||\n" (map comp4 stuff)
        val actlist = (map comp5 stuff)
        val gd_defs = PP_guard_defs mid inc chans actlist
        val ops  = PP_operations mid inc chans actlist
     in
        "SETS\n" ^ (PP_listop ";" [sets2,sets]) ^
        "\n\nVARIABLES\n" ^ vars ^
        ( PP_defs (PP_listop (";\n") [gd_defs,defs] ) ) ^
        "\n\nINVARIANT\n" ^ (PP_listop "&" [inv2,inv]) ^
        "\nINITIALISATION\n" ^ init ^
        "\nOPERATIONS\n\n" ^ ops ^
        "\n\nEND\n"
     end
;


fun PP_includes inc_id inc_params inc_list =
    let fun f (id,params) = id ^ 
                        ( if params = [] then ""
                          else "(" ^ (PP_listop "," params) ^ ")" )
        val l = map f ((inc_id,inc_params)::inc_list)
        val inc_str = PP_listop (",\n" ^ (tab 1)) l
    in
        if inc_str = "" then ""
        else "INCLUDES\n" ^ (tab 1) ^ inc_str ^ "\n\n"
    end
;


fun PP_machine fname comment mid mparams preamble inc_id inc_params
               inc_list sdi chans pinfo_list = 
    let 
        val header1 = "MACHINE " ^ mid ^ 
                    (if mparams = [] then "\n\n"
                     else "(" ^ (PP_listop "," mparams) ^ ")\n\n" ) ^ 
                    preamble ^ "\n\n"
        val header2 = PP_includes inc_id inc_params inc_list
        val rest = PP_rest mid sdi (not(inc_id="")) chans pinfo_list
    in 
        "\n/* Generated from " ^ fname ^ " by csp2B on " ^
        (DatetoString (fromTimeLocal (now()))) ^ ". */\n\n" ^
        comment ^
        header1 ^ header2 ^ rest
    end
;



fun PP_refinement fname comment mid rid preamble inc_id inc_params 
                  inc_list sdi chans pinfo_list = 
    let 
        val header1 = "REFINEMENT " ^ mid ^ "\n\n" ^
                       "REFINES " ^ rid ^ "\n\n" ^ 
                       preamble ^ "\n\n"
        val header2 = PP_includes inc_id inc_params inc_list
        val rest = PP_rest mid sdi (not(inc_id="")) chans pinfo_list
    in 
        "\n/* Generated from " ^ fname ^ " by csp2B on " ^
        (DatetoString (fromTimeLocal (now()))) ^ ". */\n\n" ^
        comment ^
        header1 ^ header2 ^ rest
    end
;






















































exception SynErr of (string*(int*int));


fun take_indices_from_idents pos eid [] ids = ids
  | take_indices_from_idents pos eid (i::ids) [] = 
        raise SynErr ("Missing index '" ^ i ^ "' for event '" ^ eid ^ "'", pos)
  | take_indices_from_idents pos eid (i::ids) (j::jds) = 
      if i = j
      then take_indices_from_idents pos eid ids jds
      else raise SynErr ("Wrong index '" ^ j ^ "' for event '" ^ eid ^ "'", pos)
;


fun take_indices_from_params pos [] params = params
  | take_indices_from_params pos (i::ids) [] = 
        raise SynErr ("Missing index '" ^ i ^ "'", pos)
  | take_indices_from_params pos (i::ids) (p::params) = 
      case p of 
        (Dot e) => if i = expr_to_string e 
                   then take_indices_from_params pos ids params
                   else raise SynErr ("Wrong indexing variable '" ^
                                     (expr_to_string e) ^ "'", pos)
         | v => raise SynErr ("Wrong event indices", pos)
;



fun 
      parse_identlist es (t,s) =
                if is_identifier t
                then parse_identlist2 (es@[t]) (lexscan s)
                else raise SynErr ("Identifier expected", get_pos(t,s))
and
      parse_identlist2 es (t,s) =
                if t = ","
                then parse_identlist es (lexscan s)
                else  (es,(t,s)) 
;



fun 
      parse_setlist es (t,s) =
                if is_identifier t
                then parse_setlist2 (es@[t]) (lexscan s)
                else raise SynErr ("Identifier expected", get_pos(t,s))
and
      parse_setlist2 es (t,s) =
                if t = ";"
                then parse_setlist es (lexscan s)
                else  (es,(t,s)) 
;


fun parse_opt_brac_identlist (t,s) =
    if t = "("
    then let val (params, (t2,s2)) = parse_identlist [] (lexscan s)
         in if t2 = ")"
            then (params,lexscan s2)
            else raise SynErr("')' expected", get_pos(t2,s2))
         end
    else ([], (t,s));



fun parse_index_list indices1 (t,s) =
    if t = "["
    then let val r = (lexscan s)
             val pos = get_pos(r)
             val (indices2, (t2,s2)) = parse_identlist [] r
         in if t2 = "]"
            then if indices1 = indices2
                 then (indices2,lexscan s2)
                 else raise SynErr("Wrong indexing variables", pos)
            else raise SynErr("']' expected", get_pos(t2,s2))
         end
    else if indices1 = []
         then ([], (t,s))
         else raise SynErr("Wrong indexing variables", get_pos(t,s))
;








fun parse_expr1 endings (b,c,d) e (t,s) =
        if (t mem endings  andalso (b,c,d) = (0,0,0) )
           orelse t=""
        then (e, (t,s))
        else  let val (b1,c1,d1) =
                case t of
                "("   => (b+1, c  , d  )
              | "{"   => (b  , c+1, d  )
              | "["   => (b  , c  , d+1)
              | ")"   => (b-1, c  , d  )
              | "}"   => (b  , c-1, d  )
              | "]"   => (b  , c  , d-1)
              |  v    => (b  , c  , d  )   in
           parse_expr1 endings (b1, c1, d1) (e@[t]) (lexscan s)  end;


fun parse_expr endings (t,s) = 
        let val (e,(t2,s2)) = parse_expr1 endings (0,0,0) [] (t,s) in
                if e = []
                then raise SynErr ("expression expected", get_pos(t,s))
                else (e, (t2,s2) ) end;




fun 
      parse_exprlist endings es r =
        let val (e,(t2,s2)) = parse_expr endings r
        in
                if t2 = ","
                then parse_exprlist endings (es@[e]) (lexscan s2)
                else  ( es@[e], (t2,s2) )
        end
;


fun parse_opt_brac_exprlist (t,s) =
    if t = "("
    then let val (params, (t2,s2)) = parse_exprlist [")",","] [] (lexscan s)
         in if t2 = ")"
            then (params,lexscan s2)
            else raise SynErr("')' expected", get_pos(t2,s2))
         end
    else ([], (t,s))
;





fun parse_recurse env (id,s) =   
    let val r2 = lexscan s
        val pos = get_pos(r2)
        val (indices,r3) = parse_index_list (get_indices env) r2
        val (args,r4) = parse_opt_brac_exprlist r3
    in
            (MK_recurse id args, r4)
    end
;



fun parse_event_inparams env pos params (t,s) = 
  let val endings = [ "->", "?", "!", "." ]
  in
    case t of
        "?" => let val (id,s2) = lexscan s 
               in  parse_event_inparams env pos (params @ [MK_inparam id]) (lexscan s2)  
               end 
      | "." => let val (e,r2) = parse_expr endings (lexscan s)
               in  parse_event_inparams env pos (params @ [MK_dotparam e]) r2
               end
      |  v  => ( take_indices_from_params pos (get_indices env) params , (t,s))
  end
;


fun parse_event_outparams params (t,s) = 
  let val endings = [ "->",  "!"  ]
  in
    if t = "!"
      then let val (e,r2) = parse_expr endings (lexscan s)
               in  parse_event_outparams (params @ [MK_outparam e]) r2
               end
      else  ( params , (t,s))
  end
;



fun parse_event_params env eid pos params r = 
  let 
    val ( ins, r2 ) =  parse_event_inparams env pos [] r
    val ( outs, r3 ) =  parse_event_outparams [] r2
    val (l1,l2) = get_param_lengths env eid
  in
    if length(ins) = l1
    then if length(outs) = l2
         then ( ins@outs, r3 )
         else raise SynErr( "Wrong number of output parameters", pos )
    else raise SynErr( "Wrong number of input parameters", pos )
  end
;




fun
        parse_proclist procs env (t,s) =
                let val (p,(t2,s2)) = parse_proc env (t,s)
                in if t2 = "[]"
                   then parse_proclist (procs@[p]) env (lexscan s2)
                   else (MK_sum (procs@[p]), (t2,s2))
                end
and                
        parse_proc env (t,s) =
                case t of
                          "STOP"        => (MK_stop , lexscan s)
                        | "IF"          => parse_if env (lexscan s)
                        | "("           => parse_br_proclist env (lexscan s)
                        | v             => 
                            if is_eventname env t
                            then parse_prefix env (t,s)
                            else raise SynErr(("Event name expected") , get_pos(t,s))
and                
        parse_uproc env (t,s) =
                case t of
                          "STOP"        => (MK_stop , lexscan s)
                        | "IF"          => parse_if env (lexscan s)
                        | "("           => parse_br_proclist env (lexscan s)
                        | v             => 
                            if is_eventname env t
                            then parse_prefix env (t,s)
                            else  if is_identifier t
                            then parse_recurse env (t,s)
                            else raise SynErr(("Illegal term '" ^ t ^ "'") , get_pos(t,s))
and
        parse_if env r = 
            let val (g,(t2,s2)) = parse_expr ["THEN"] r
            in
                if t2 = "THEN"
                then let val (p,(t3,s3)) = parse_proc env (lexscan s2)
                     in
                        if t3 = "ELSE"
                        then let val (q,r4) = parse_proc env (lexscan s3)
                             in
                                (MK_ifelse g p q, r4)
                             end
                        else (MK_if g p, (t3,s3))
                     end
                else raise SynErr("'THEN' expected", get_pos(t2,s2))
            end
and
        parse_br_proclist env (t,s) =
            let val (p,(t2,s2)) = parse_proclist [] env (t,s)
            in if t2 = ")"
               then (p, lexscan s2)
               else raise SynErr("')' expected", get_pos(t2,s2))
            end
and
        parse_prefix env (id,s) =  (* is_identifier(id) *)
            let val r5 = (lexscan s)
                val pos = get_pos (id,s)
                 val (params,(t2,s2)) = parse_event_params env id pos [] r5
            in  if t2 = "->"
                then let val (p,r3) = parse_uproc env (lexscan s2)
                     in  (MK_prefix (MK_event id params) p, r3)
                     end
                else raise SynErr("'->' expected", get_pos(t2,s2))
            end
;






fun replicate x i = 
        if i=0
        then []
        else x::(replicate x (i-1));



fun 
  parse_dec endings (t,s) =
      if is_identifier(t)
      then
        let val (t2,s2) = lexscan s in
                if t2 = ":"
                then let val (el,(t3,s3)) = parse_expr endings (lexscan s2) 
                         val e = expr_to_string el
                     in
                        ((t, e), (t3,s3) ) end
                else raise SynErr ("':' expected", get_pos(t2,s2))
        end
      else raise SynErr ("Identifier expected", get_pos(t,s))
;


fun
  parse_declist endings r =
        let val ((var,ty),r2) = parse_dec endings r 
            val ((vars,tys),r3) = parse_declist2 endings r2
        in
           ((var::vars, ty::tys), r3)
        end
and
  parse_declist2 endings (t,s) =
        if t = ","
        then parse_declist endings  (lexscan s)
        else  (([],[]),(t,s)) 
;



fun parse_opt_brac_declist (t,s) =
    if t = "("
    then let val (params, (t2,s2)) = parse_declist [",",")"] (lexscan s)
         in if t2 = ")"
            then (params,lexscan s2)
            else raise SynErr("')' expected", get_pos(t2,s2))
         end
    else (([],[]), (t,s));



fun 
  parse_opt_dec endings (t,s) =
      if is_identifier(t)
      then
        let val (t2,s2) = lexscan s in
                if t2 = ":"
                then let val (el,(t3,s3)) = parse_expr endings (lexscan s2) 
                         val e = expr_to_string el
                     in
                        ((t, e), (t3,s3) ) end
                else ((t, "?"), (t2,s2) )
        end
      else raise SynErr ("Identifier expected", get_pos(t,s))
;


fun
  parse_opt_declist endings r =
        let val ((var,ty),r2) = parse_opt_dec endings r 
            val ((vars,tys),r3) = parse_opt_declist2 endings r2
        in
           ((var::vars, ty::tys), r3)
        end
and
  parse_opt_declist2 endings (t,s) =
        if t = ","
        then parse_opt_declist endings  (lexscan s)
        else  (([],[]),(t,s)) 
;



fun parse_opt_brac_paramlist (t,s) =
    if t = "("
    then let val (params, (t2,s2)) = parse_opt_declist [",",")"] (lexscan s)
         in if t2 = ")"
            then (params,lexscan s2)
            else raise SynErr("')' expected", get_pos(t2,s2))
         end
    else (([],[]), (t,s));




fun parse_ident (t,s) = 
    if is_identifier(t) 
    then (t,lexscan s) 
    else raise raise SynErr("Identifier expected", get_pos(t,s))
;



fun parse_keyword w (t,s) =
    if t = w
    then (lexscan s)
    else raise SynErr("'" ^ w ^ "' expected", get_pos(t,s))
;



fun parse_initproc2 (t,s) =
    let val (indices,(t3,s3)) = 
                if t = "|||" 
                then  let val (indices2,(t2,s2)) =
                            parse_identlist [] (lexscan s)
                      in if t2 = "."
                         then (indices2, lexscan s2)
                         else raise SynErr("'.' expected", get_pos(t2,s2))
                      end
                else ( [], (t,s) )
    in
        if is_identifier(t3)
        then let 
                val r5 = (lexscan s3)
                val pos = get_pos r5
                val (indices3,r6) = parse_index_list indices r5
                val (args,r7) = parse_opt_brac_exprlist r6
             in
                (indices, t3, args, r7)
             end
        else raise SynErr("Process identifier expected", get_pos(t3,s3))
     end
;


fun parse_initproc (t,s) =
    if is_identifier(t)
    then let val (t2,s2) = lexscan s
         in if t2 = "="
            then let val (indices,initpid,args,r3) =parse_initproc2 (lexscan s2)
                 in (t,indices,initpid,args,r3) end
            else raise SynErr("'=' expected", get_pos(t2,s2))
         end
    else  raise SynErr("Process identifier expected", get_pos(t,s))
;


fun check_fresh_ids pos used [] = true
  | check_fresh_ids pos used (i::t) =
        if member(used,i)
        then raise SynErr( "Variable '" ^ i ^ "' already in use", pos)
        else check_fresh_ids pos used t
;


fun check_no_dups pos vars =
    if  length vars =  length (rem_dups [] vars) 
    then ()
    else raise SynErr( "Variable list contains duplicates", pos )
;


fun parse_eqn env used (t,s) =
    if is_identifier t
    then let val _ = check_fresh_ids (get_pos(t,s)) used [t]
             val r2 = (lexscan s)
             val pos = get_pos r2
             val (indices2,r3) = parse_index_list (get_indices env) r2
             val ((vars,types),(t4,s4)) = parse_opt_brac_paramlist r3
             val _ = check_fresh_ids pos used vars
             val _ = check_no_dups pos vars
             val env2 = MK_update_types env (vars,types)
         in if t4 = "="
            then let val (p,r5) = parse_proclist [] env (lexscan s4)
                 in ( t, env2, MK_eqn (MK_lhs t vars) p, r5 )
                 end
            else raise SynErr("'=' expected", get_pos(t4,s4))
         end
    else raise SynErr("Identifier expected", get_pos(t,s))
;



fun parse_eqnlist env used eqs (t,s) =
    if t = "END"
    then (used, env, eqs, lexscan s)
    else let val (pid, env2, eq, r2) = parse_eqn env used (t,s)
             val used2 = add(used,pid)
         in parse_eqnlist env2 used2 (eqs@[eq]) r2
         end
;


fun check_lists st pos l1 [] = ()
  | check_lists st pos l1 (x::t) =
    if x mem l1
    then check_lists st pos l1 t
    else raise SynErr("Undeclared " ^ st ^ " parameter '" ^ x ^ "'", pos)
;

fun check_channel_constrains cid cidpos outs outpos ins inpos [] =
        raise SynErr("Undeclared event '" ^ cid ^ "'" , cidpos)
  | check_channel_constrains cid cidpos outs outpos ins inpos 
            ( (cid2,(ins2,_),(outs2,_))::chans )  =
        if cid = cid2
        then ( check_lists "output" outpos outs2 outs ;
               check_lists "input" inpos ins2 ins  )
        else check_channel_constrains cid cidpos outs outpos ins inpos chans
;
    

fun parse_channels_constrains_out (t,s) =
    if t = "("
    then let val (outs,(t2,r2)) = parse_opt_brac_identlist (t,s)
         in  if t2="<--"
             then (outs,lexscan r2)
             else raise SynErr("'<--' expected", get_pos(t2,r2))
         end
    else
        let val (t2,s2) =lexscan s
        in
            if t2 = "<--" andalso is_identifier(t)
            then ( [t], lexscan s2)
            else ( [], (t,s) )
        end
;
                   

fun parse_channels_constrains3 chans indices (t,s) =
    if t = "WHERE"
    then ([],(t,s))
    else let
            val outpos = get_pos(t,s)
            val (outs,(cid,s2)) = parse_channels_constrains_out (t,s)
            val cidpos = get_pos(cid,s2)
         in
            if is_identifier(cid)
            then
                let val r6 = (lexscan s2)
                    val inpos = get_pos r6
                    val (ins1,r3) = parse_opt_brac_identlist r6
                    val _ = check_channel_constrains  cid cidpos
                                            outs outpos ins1 inpos chans
                    val ins2 = take_indices_from_idents inpos cid indices ins1
                    val (rest,r4) = parse_channels_constrains3 chans indices r3
                in
                    ( (cid,ins2,outs)::rest, r4 )
                end
            else raise SynErr("Operation name expected", get_pos(cid,s2))
         end
;


fun parse_channels_constrains2 chans indices (t,s) =
         let
            val outpos = get_pos(t,s)
            val (outs,(cid,s2)) = parse_channels_constrains_out (t,s)
            val cidpos = get_pos(cid,s2)
         in
            if is_identifier(cid)
            then
                let val r6 = (lexscan s2)
                    val inpos = get_pos r6
                    val (ins1,r3) = parse_opt_brac_identlist r6
                    val _ = check_channel_constrains  cid cidpos
                                            outs outpos ins1 inpos chans
                    val ins2 = take_indices_from_idents inpos cid indices ins1
                    val (rest,r4) = parse_channels_constrains3 chans indices r3
                in
                    ( (cid,ins2,outs)::rest, r4 )
                end
            else raise SynErr("Operation name expected", get_pos(cid,s2))
         end
;



fun check_indices_chans pos eid [] [] = ()
  | check_indices_chans pos eid (x::t) [] = ()
  | check_indices_chans pos eid [] (y::s) = 
        raise SynErr("Missing index '" ^ y ^ "' for event '" ^ eid ^"'",pos)
  | check_indices_chans pos eid (x::t) (y::s) = 
        if x = y then check_indices_chans pos eid t s
        else raise SynErr("Wrong index '" ^ x ^ "' for event '" ^ eid ^"'",pos)
;

fun parse_channels_constrains chans indices (t,s) =
    if t = "CONSTRAINS" 
    then parse_channels_constrains2 chans indices (lexscan s)
    else let val pos = get_pos(t,s)
             fun f (x,(ins,_),(outs,_)) = 
                    let val  _ = check_indices_chans pos x ins indices
                    in (x,  drop(ins,length indices), outs ) end
         in   ( map f chans, (t,s))
         end
;


fun check_no_dups_chan_constrains pos ids =
    if  length ids =  length (rem_dups [] ids) 
    then ()
    else raise SynErr( "Operation names are duplicated", pos )
;

fun fst3 (x,_,_) = x;


fun parse_procdef chans chantypes fresh used s = 
         let val r = (lexscan s)
             val spos = get_pos r
             val (pid,indices,initpid,initargs,r2) = parse_initproc r
             val _ = if member(used,pid)
                     then raise SynErr("Identifier '" ^ pid ^ "' already in use",spos)
                     else ()
             val used5 = add(used,pid)
             val pos = get_pos r2
             val (eventdecs,r3) = parse_channels_constrains chans indices r2
             val _ = check_no_dups_chan_constrains pos 
                        (map fst3 eventdecs)
             val env2 = MK_env chans chantypes indices eventdecs
             val r4 = parse_keyword "WHERE" r3
             val (used2,env3,eqs,r5) =  parse_eqnlist env2 used5 [] r4
             val (fresh2,env4,eqs2) = eqnsnorm env3 fresh eqs
             val used3 = addList(used2,(get_var_ids env4))
          in  
             ( fresh2, used3 , 
               (pid,initpid,initargs,eventdecs,indices,get_types env4,eqs2),  r5 )
          end
;



fun parse_procdef_list chans chantypes used fresh  (t,s) =
    if t = "PROCESS"
    then let val (fresh2,used2,pinfo,r2) =
                    parse_procdef chans chantypes fresh used s
             val (pinfo_list,r3) =
                    parse_procdef_list chans chantypes used2 fresh2 r2
         in  ( pinfo::pinfo_list, r3)
         end
    else if t="END"
    then ([],(lexscan s))
    else raise SynErr("'END' expected", get_pos(t,s))
;


fun parse_includes2 r =
    let val (id,r2) = parse_ident r
        val (params, (t3,s3)) = parse_opt_brac_exprlist r2
        val pitem = (id, map expr_to_string params)
    in
        if t3 = ","
        then let val (plist,r4) = parse_includes2 (lexscan s3)
             in
                ( pitem::plist, r4)
             end
        else ( [pitem], (t3,s3) )
    end
;
        

fun parse_includes (t,s) =
    if t = "INCLUDES"
    then parse_includes2 (lexscan s)
    else ([], (t,s))
;


        

fun parse_conjoins (t,s) =
    if t = "CONJOINS"
    then let val (id,r2) = parse_ident (lexscan s)
             val (params, r3) = parse_opt_brac_exprlist r2
         in
             (id, map expr_to_string params, r3)
         end
    else ("",[], (t,s))
;


fun parse_preamble (t,s) =
    if t mem ["USES", "SEES", "IMPORTS", "EXTENDS", "PROMOTES", "CONSTRAINTS" ]
    then let val (p,r) =
           lexscan_until ["INCLUDES", "CONJOINS", "SETS", 
                          "DEFINITIONS", "INVARIANT", "ALPHABET"] s
         in
           ( t ^ "\n\n" ^ p, r)
         end
    else  ("",(t,s)) ;
           

fun parse_sets (t,s) =
    if t = "SETS"
    then  lexscan_until ["DEFINITIONS", "INVARIANT", "ALPHABET"] s
    else ("",(t,s))
;


fun parse_defns (t,s) =
    if t = "DEFINITIONS"
    then  lexscan_until ["INVARIANT", "ALPHABET"] s
    else ("",(t,s))
;


fun parse_inv (t,s) =
    if t = "INVARIANT"
    then  lexscan_until ["ALPHABET"] s
    else ("",(t,s))
;

fun parse_SDI r =
    let val (s,r2) = parse_sets r
        val (d,r3) = parse_defns r2
        val (i,r4) = parse_inv r3
    in
        ((s,d,i),r4)
    end
;



fun parse_channelsout (t,s) =
    if t = "("
    then let val (outs,(t2,r2)) = parse_opt_brac_paramlist (t,s)
         in  if t2="<--"
             then (outs,lexscan r2)
             else raise SynErr("'<--' expected", get_pos(t2,r2))
         end
    else
        let val (t2,s2) =lexscan s
        in
            if t2 = "<--" andalso is_identifier(t)
            then ( ([t],["?"]), lexscan s2)
            else ( ([],[]), (t,s) )
        end
;
                   

fun parse_channels2  ending (t,s) =
    if t = ending
    then ([],(t,s))
    else let
            val (outs,(cid,s2)) = parse_channelsout (t,s)
         in
            if is_identifier(cid)
            then
                let
                    val (ins,r3) = parse_opt_brac_declist (lexscan s2)
                    val (chans,r4) = parse_channels2 ending r3
                in
                    ( (cid,ins,outs)::chans, r4 )
                end
            else raise SynErr("Operation name expected", get_pos(cid,s2))
         end
;

fun parse_channels  ending (t,s) =
        let
            val (outs,(cid,s2)) = parse_channelsout (t,s)
         in
            if is_identifier(cid)
            then
                let
                    val (ins,r3) = parse_opt_brac_declist (lexscan s2)
                    val (chans,r4) = parse_channels2 ending r3
                in
                    ( (cid,ins,outs)::chans, r4 )
                end
            else raise SynErr("Operation name expected", get_pos(cid,s2))
         end
;


fun get_used_ids [] = []
  | get_used_ids ((eid,(ins,_),(outs,_))::t) =
        (eid::ins) @ outs @ (get_used_ids t)
;


fun parse_machine fname comment r =
    let val (mid, r2) =  parse_ident r
        val (mparams, r2a) = parse_opt_brac_identlist r2
        val (preamble, r2b) = parse_preamble r2a
        val (inc_list, r3) = parse_includes r2b
        val (inc_id,inc_params,r4) = parse_conjoins r3
        val (sdi,r10) = parse_SDI r4
        val r5 = parse_keyword "ALPHABET" r10
        val ( chans, r6 ) =  parse_channels "PROCESS" r5
        val chantypes = MK_chantypes chans
        val used = addList ( empty scompare, 
                             (get_used_ids chans) )
        val ( pinfo_list, r7) = 
                parse_procdef_list chans chantypes used MK_empty_fresh r6
        val result = PP_machine fname comment mid mparams preamble
                                inc_id inc_params inc_list
                                sdi chans pinfo_list
        val outfile = openOut (mid ^ ".mch")
     in

        output (outfile,  result ) ;
        closeOut outfile;   
        print ("\nTranslation finished.\nResults placed in '" 
                              ^ mid ^ ".mch'.\n\n")
     end
;



fun parse_refinement fname comment r =
    let val (mid, r2) =  parse_ident r
        val r3 = parse_keyword "REFINES" r2
        val (rid,r3a) = parse_ident r3
        val (preamble, r3b) = parse_preamble r3a
        val (inc_list, r4) = parse_includes r3b
        val (inc_id,inc_params,r5) = parse_conjoins r4
        val (sdi,r6) = parse_SDI r5
        val r7 = parse_keyword "ALPHABET" r6
        val ( chans, r8 ) =  parse_channels "PROCESS" r7
        val chantypes = MK_chantypes chans
        val used = addList ( empty scompare, (get_used_ids chans) )
        val ( pinfo_list, r9) = 
                parse_procdef_list chans chantypes used MK_empty_fresh r8
        val result = PP_refinement fname comment mid rid preamble
                                    inc_id inc_params inc_list
                                    sdi chans pinfo_list
        val outfile = openOut (mid ^ ".ref")
     in

        output (outfile,  result ) ;
        closeOut outfile;   
        print ("\nTranslation finished.\nResults placed in '" 
                              ^ mid ^ ".ref'.\n\n")
     end
;



fun csp2B filename = 
  let 
    val infile = openIn (filename)
    val st1 = inputAll infile 
    val _ = closeIn infile
    val (comment, (t,s)) = lexscan_until ["MACHINE", "REFINEMENT"] 
                        (init_string2 st1)
  in
     case t of
        "MACHINE" => parse_machine filename comment (lexscan s)
      | "REFINEMENT" => parse_refinement filename comment (lexscan s)
      | v => raise SynErr("'MACHINE' or 'REFINEMENT' expected", get_pos(t,s))
  end 
  handle SynErr(msg,(i,j)) =>
          (print ("\nError: " ^ msg ^ " at line " ^ 
                        (makestring i) ^
                        " position " ^ (makestring j) ^ ".\n" ^
                        "Translation failed.\n\n")  
          )
       | Io{cause=_, function=_, name=s} => 
          print ("\nFile '" ^ s ^ "' not found.\n\n")
   ; 


val args = arguments();

val _ = 
    (if length(args) = 1
    then csp2B (hd args)
    else print "\nUsage: csp2B filename.\n\n")
    ;


end
